import numpy as np

from .mrf import neighbors

def fwd_bwd_sched(mrf):
  """Constructs a forward-backward message update schedule.

  It's assumed that the nodes are in a chain in the order they appear in
  `mrf.nodes`.

  Parameters
  ----------
  mrf : MRF

  Returns
  -------
  Message update schedule.
  """
  return zip(mrf.nodes, mrf.nodes[1:]) \
       + zip(reversed(mrf.nodes), reversed(mrf.nodes[:-1]))

def full_sched(mrf):
  """Returns a message update schedule that updates each edge in no particular
  order.

  Parameters
  ----------
  mrf : MRF

  Returns
  -------
  A schedule which covers ever edge in every direction.
  """
  return [(s, t) for (s, t) in mrf.edges] + [(t, s) for (s, t) in mrf.edges]

class MessagePassingScheme(object):
  def __init__(self, mrf):
    self.mrf = mrf

  def messages(self, node_pot, edge_pot):
    pass

  def log_beliefs(self, node_pot, edge_pot):
    pass

class MaxSumBP(MessagePassingScheme):
  def __init__(self,
               mrf,
               max_iters=100,
               conv_tol=1e-5,
               stepsize=1.0,
               sched=None):
    """Max-product belief propogation (max sum over log potentials).

    Parameters
    ----------
    mrf : MRF
    max_iters : int (default: 100)
    conv_tol : float (default: 1e-5)
        The message passing algorithm will abort once the maximum difference
        between messages across iterations is less than `conv_tol`.
    stepsize : float (default: 1.0)
        A factor which controls the "strength" of updates to messages. Use 1.0
        for tree-structured graphs.
    sched : list of edges or None
        The schedule by which messages should be sent. Defaults to
        `full_sched(mrf)`.
    """
    super(self.__class__, self).__init__(mrf)
    self.max_iters = max_iters
    self.conv_tol = conv_tol
    self.stepsize = stepsize
    self.sched = full_sched(mrf) if sched == None else sched

  def message_foundation(self, node_pot, edge_pot, msg, nStates, s, t):
    # Get the unary potential at s (nStates[s])
    pot_s = node_pot[s]

    # Get edge potential between s and t (nStates[s] x nStates[t])
    pot_st = edge_pot[(s, t)] if (s, t) in edge_pot else edge_pot[(t, s)].T

    # Compute incoming messages to s except for t (nStates[s])
    incoming_msg = sum([msg[(v, s)] for v in neighbors(self.mrf, s) if v != t],
                       np.zeros(nStates[s]))

    # Return the total message matrix (nStates[s] x nStates[t])
    return pot_s.reshape((nStates[s], 1)) \
         + pot_st \
         + incoming_msg.reshape((nStates[s], 1))

  def messages(self, node_pot, edge_pot):
    nStates = {v: len(node_pot[v]) for v in self.mrf.nodes}

    # Init old messages to 0 everywhere
    msg_old = {}
    for (s, t) in self.mrf.edges:
      msg_old[(s, t)] = np.zeros(nStates[t])
      msg_old[(t, s)] = np.zeros(nStates[s])

    # Init messages to be uniform
    msg = {}
    for (s, t) in self.mrf.edges:
      msg[(s, t)] = np.log(1.0 / nStates[t]) * np.ones(nStates[t])
      msg[(t, s)] = np.log(1.0 / nStates[s]) * np.ones(nStates[s])

    stats = {'last_iter': None, 'error': [], 'converged': False}

    for it in range(self.max_iters):
      # Set the damping factor
      damp = 1.0 if (it == 0) else self.stepsize

      for (s, t) in self.sched:
        # Calculate message s => t
        msg_foundation = self.message_foundation(node_pot, edge_pot, msg, \
            nStates, s, t)
        new_msg = np.max(msg_foundation, axis=0)
        # new_msg_argmax = np.argmax(msg_mat, axis=0)

        # Normalize for numerical stability
        new_msg -= np.max(new_msg)

        # Damp and set messages
        new_msg_damped = damp * new_msg + (1 - damp) * msg_old[(s, t)]
        msg_old[(s, t)] = np.copy(msg[(s, t)])
        msg[(s, t)] = new_msg_damped

      # Check convergence
      edge_error = lambda s, t: \
          np.max(np.abs(np.exp(msg[(s, t)]) - np.exp(msg_old[(s, t)])))
      # Add zero on the end to ensure that the list is non-empty and will
      # default to zero.
      msg_diff = max([max(edge_error(s, t), edge_error(t, s))
                      for (s, t) in self.mrf.edges] + [0.0])
      stats['error'].append(msg_diff)

      if it > 0 and msg_diff < self.conv_tol:
        # We have converged!
        stats['converged'] = True
        stats['last_iter'] = it
        break

    return msg, stats

  def log_beliefs(self, node_pot, edge_pot, msg):
    nStates = {v: len(node_pot[v]) for v in self.mrf.nodes}

    node_bel = {}
    for v in self.mrf.nodes:
      # Note that we start the sum with the node potentials.
      pre_msg = sum([msg[(t, v)] for t in neighbors(self.mrf, v)], node_pot[v])
      node_bel[v] = pre_msg - np.max(pre_msg)

    return node_bel

  def decode_MAP_states(self, node_pot, edge_pot, node_bel):
    map_state = {v: np.argmax(node_bel[v]) for v in self.mrf.nodes}
    n_ties = sum([np.sum(np.abs(node_bel[v]
                    - node_bel[v][map_state[v]]) < self.conv_tol) - 1
                  for v in self.mrf.nodes])
    return map_state, n_ties



import numpy as np
import random

def random_walk_proposal(cov):
  """Random walk proposal with covariance `cov`.

  Parameters
  ----------
  cov : d x d matrix

  Returns
  -------
  proposal : function
      A proposal function which randomly selects points and adds Gaussian random
      noise to them with covariance `cov`.
  """
  cov = np.array(cov)
  d = cov.shape[0]
  mu = np.zeros(d)

  def proposal(x, mrf, nAdd):
    x_prop = {}
    for v in mrf.nodes:
      x_prop[v] = [np.random.multivariate_normal(mu, cov) + random.choice(x[v])
                   for _ in range(nAdd[v])]
    return x_prop

  return proposal

def random_walk_proposal_1d(sig):
  """Random walk proposal for 1-dimensional models.

  Parameters
  ----------
  sig : float
      The standard deviation of the random walk.

  Returns
  -------
  proposal : function
      A random walk proposal function.
  """
  def proposal(x, mrf, nAdd):
    x_prop = {}
    for v in mrf.nodes:
      x_prop[v] = [sig * np.random.randn() + random.choice(x[v])
                   for _ in range(nAdd[v])]
    return x_prop

  return proposal



import numpy as np

class MRF(object):
  def __init__(self, nodes, edges, node_pot, edge_pot):
    """Pairwise Markov Random Field.

    Parameters
    ----------
    nodes : list of vertex IDs, not necessarily integers
    edges : list of ordered pairs (s, t)
        Should only contain (s, t) or (t, s) but not both.
    node_pot : function (s, x_s -> R)
        Function for evaluating log unary potentials.
    edge_pot : function (s, t, x_s, x_t -> R)
        Function for evaluating log pairwise potentials.
    """
    self.nodes = nodes
    self.edges = edges

    # Single lambdas for each type of potential, since throwing around lambdas
    # everywhere isn't efficient.
    self.node_pot = node_pot
    self.edge_pot = edge_pot

def neighbors(mrf, v):
  """Get the neighbors of a given vertex.

  Parameters
  ----------
  v : node

  Returns
  -------
  List of nodes that are connected to v.
  """
  return [t for t in mrf.nodes if ((v, t) in mrf.edges)
                               or ((t, v) in mrf.edges)]

def calc_potentials(mrf, x):
  """Calculate all unary and pairwise log potentials.

  Parameters
  ----------
  mrf : MRF
  x : dict (v -> list of particles)
      The particle set to evaluate.

  Returns
  -------
  node_pot : dict (s -> [pot])
      Log unary potentials.
  edge_pot : dict ((s,t) -> [[pot]])
      Matrix of log pairwise potentials.
  """
  node_pot = {}
  edge_pot = {}

  for s in mrf.nodes:
    node_pot[s] = np.array([mrf.node_pot(s, x_s) for x_s in x[s]])

  for (s, t) in mrf.edges:
    pot_st = np.zeros((len(x[s]), len(x[t])))
    for (n_s, x_s) in enumerate(x[s]):
      for (n_t, x_t) in enumerate(x[t]):
        pot_st[n_s, n_t] = mrf.edge_pot(s, t, x_s, x_t)
    edge_pot[(s, t)] = pot_st

  return node_pot, edge_pot

def log_prob_states(mrf, states, node_pot, edge_pot):
  """Evaluate the log probability of a particular state sequence.

  Parameters
  ----------
  mrf : MRF
  states : dict (v -> int)
      A representation of the state of every node.
  node_pot : dict (v -> array of unary potentials)
      The log unary potentials for each particle at each node.
  edge_pot : dict ((s, t) -> Ns x Nt matrix of pairwise potentials)
      The log pairwise potentials for every pair of particles across each edge.

  Returns
  -------
  The log probability of the given state assignment.
  """
  return sum(node_pot[v][states[v]] for v in mrf.nodes) \
       + sum(edge_pot[(s, t)][states[s], states[t]] for (s, t) in mrf.edges)



import numpy as np
import random

from functools import wraps

def set_seed(seed=0):
  np.random.seed(seed)
  random.seed(seed)

def seeded(f, seed=0):
  @wraps(f)
  def f_seeded(*args, **kwargs):
    set_seed(seed)
    return f(*args, **kwargs)
  return f_seeded



from .mrf import calc_potentials, log_prob_states

def DPMP_infer(mrf,
               x0,
               nParticles,
               proposal,
               particle_selection,
               msg_passing,
               max_iters=100,
               conv_tol=1e-5,
               nAugmented=None,
               callback=None,
               temp=1.0,
               verbose=False):
  """Run D-PMP inference.

  Parameters
  ----------
  mrf : MRF
  x0 : dict (v -> list of particles)
      The initial particle set.
  nParticles : int or dict (v -> int)
      The number of particles to keep after selection each iteration.
  proposal : function (x, mrf, nAdd -> list of particles)
  particle_selection : ParticleSelectionScheme
  msg_passing : MessagePassingScheme
  max_iters : int (default: 100)
  conv_tol : float (default: 1e-5)
      The log-probability convergence tolerance. If the log-probability of the
      MAP decoding does not improve by more than conv_tol, then we assume
      convergence and stop.
  nAugmented : int or dict (v -> int) or None (default: None)
      The number of particles to be selected from at every iteration. At each
      node, nAugmented determines the number of extra particles that will be
      proposed. If None, then it defaults to twice nParticles.
  callback : function or None (default: None)
      The function to be called at the end of each iteration.
  temp : float (default: 1.0)
      Tricky. TODO.
  verbose : boolean (default: False)

  Returns
  -------
  xMAP : dict (v -> state)
      The MAP decoding found.
  x : dict (v -> list of particles)
      The final particle set.
  stats : dict
      Contains
        - 'logP', a list of the log probabilities of the MAP configurations at
          each iteration.
        - 'n_ties', the number of message passing ties at each iteration.
        - 'msg_passing_stats', the message passing stats at each iteration.
        - 'callback_results', the result of the callback function, if it exists,
          at each iteration.
        - 'converged', a boolean specifying whether DPMP converged.
        - 'last_iter', the number of iterations until convergence, if DPMP
          converged.
  """

  # Handle default arguments
  if isinstance(nParticles, int):
    nParticles = {v: nParticles for v in mrf.nodes}
  if nAugmented == None:
    nAugmented = {v: 2 * nParticles[v] for v in mrf.nodes}
  elif isinstance(nAugmented, int):
    nAugmented = {v: nAugmented * nParticles[v] for v in mrf.nodes}

  # Start with initial particle set
  x = x0

  stats = {
    'logP': [],
    'n_ties': [],
    'msg_passing_stats': [],
    'callback_results': [],
    'converged': False,
    'last_iter': None
  }

  for i in range(max_iters):
    if verbose: print('Iter', i)

    # Sample new particles
    x_aug = None
    x_prop = None
    if i > 0:
      # Propose new particles
      if verbose: print('    ... Proposing new particles')
      nParticlesAdd = {v: nAugmented[v] - len(x[v]) for v in mrf.nodes}
      x_prop = proposal(x, mrf, nParticlesAdd)

      # Construct augmented particle set
      x_aug = {v: x[v] + list(x_prop[v]) for v in mrf.nodes}
    else:
      x_aug = x

    # Calculate potentials on the augmented particle set
    if verbose: print('    ... Calculating potentials and MAP')
    if verbose: print('        ... potentials')
    node_pot, edge_pot = calc_potentials(mrf, x_aug)

    # Calculate messages, log beliefs, and MAP states
    if verbose: print('        ... message passing')
    msgs, msg_passing_stats = msg_passing.messages(node_pot, edge_pot)
    node_bel_aug = msg_passing.log_beliefs(node_pot, edge_pot, msgs)
    map_states, n_ties = msg_passing.decode_MAP_states(node_pot, edge_pot, \
        node_bel_aug)
    logP_map = log_prob_states(mrf, map_states, node_pot, edge_pot)
    xMAP = {v: x_aug[v][map_states[v]] for v in mrf.nodes}

    stats['logP'].append(logP_map)
    stats['n_ties'].append(n_ties)
    stats['msg_passing_stats'].append(msg_passing_stats)

    # Particle selection
    if verbose: print('    ... Selecting particles')
    accept_idx = particle_selection.select(mrf, map_states, msg_passing, msgs, \
        x_aug, nParticles, node_pot, edge_pot, temp)
    x_sel = {v: [x_aug[v][i] for i in accept_idx[v]] for v in mrf.nodes}

    # Set particles to be the selected particle set
    x = x_sel

    # Callback
    if callback != None:
      cb_res = callback({
        'mrf': mrf,
        'x0': x0,
        'nParticles': nParticles,
        'proposal': proposal,
        'particle_selection': particle_selection,
        'msg_passing': msg_passing,
        'max_iters': max_iters,
        'conv_tol': conv_tol,
        'nAugmented': nAugmented,
        'temp': temp,
        'verbose': verbose,

        'iter': i,
        'x': x,
        'x_prop': x_prop,
        'x_aug': x_aug,
        'x_sel': x_sel,
        'xMAP': xMAP,

        'stats': stats
      })
      stats['callback_results'].append(cb_res)

    # If the difference in logP between this iteration and the previous is less
    # than conv_tol, then we have converged
    if i > 0 and abs(stats['logP'][-1] - stats['logP'][-2]) < conv_tol:
      stats['converged'] = True
      stats['last_iter'] = i
      break

  # Run final message passing
  node_pot, edge_pot = calc_potentials(mrf, x)

  msgs, msg_passing_stats = msg_passing.messages(node_pot, edge_pot)
  node_bel_aug = msg_passing.log_beliefs(node_pot, edge_pot, msgs)
  map_states, n_ties = msg_passing.decode_MAP_states(node_pot, edge_pot, \
      node_bel_aug)
  xMAP = {v: x[v][map_states[v]] for v in mrf.nodes}

  return xMAP, x, stats



import numpy as np

from .mrf import neighbors

class ParticleSelectionScheme(object):
  def select(self, mrf, map_state, msg_passing, msg, x_aug, nSelect, node_pot, \
      edge_pot, temp):
    pass

class SelectDiverse(ParticleSelectionScheme):
  def select(self, mrf, map_state, msg_passing, msg, x_aug, nSelect, node_pot, \
      edge_pot, temp):
    """Run particle selection from the ICML 2014 paper (dpmpmax)."""
    nStates = {v: len(node_pot[v]) for v in mrf.nodes}

    I_accept = {v: [] for v in mrf.nodes}
    for t in mrf.nodes:
      # Build the "stacked" message foundations matrix (S x T)
      M_t = np.vstack([msg_passing.message_foundation(node_pot, edge_pot, msg, \
                           nStates, t, s).T
                       for s in neighbors(mrf, t)])

      logMstar = np.max(M_t, axis=1)                        # (S,)
      logZ = 1.0 / temp * np.max(M_t)                       # scalar
      Mstar = np.exp(1.0 / temp * logMstar - logZ)          # (S,)
      Psi = np.exp(1.0 / temp * M_t - logZ)                 # (S, T)

      # Select first particle
      b = map_state[t]
      Mhat = Psi[:,b]                                       # (S,)
      delta = Mstar - Mhat                                  # (S,)
      I_accept[t].append(b)

      while len(I_accept[t]) < nSelect[t] and np.max(delta) > 0:
        # Get unused particles
        b_used = set(I_accept[t])
        b_unused = [bb for bb in range(nStates[t]) if bb not in b_used]

        # Select next particle
        a_star = np.argmax(delta)
        min_a = delta[a_star]
        idx_max = np.argmax(Psi[a_star, b_unused])
        b = b_unused[idx_max]
        I_accept[t].append(b)

        # Update message approximation
        Mhat = np.maximum(Mhat, Psi[:,b])
        delta = Mstar - Mhat
    return I_accept

class SelectLazyGreedy(ParticleSelectionScheme):
  pass



from distutils.core import setup

setup(
  name='pyDPMP',
  version='0.1.2',
  description='The Diverse Particle Max-Product algorithm (D-PMP)',
  author='Samuel Ainsworth',
  author_email='skainsworth@gmail.com',
  url='https://github.com/samuela/pyDPMP',
  packages=['pyDPMP'],
  install_requires=['numpy']
)
